/* abc - The AspectBench Compiler
 * Copyright (C) 2007 Eric Bodden
 *
 * This compiler is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This compiler is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this compiler, in the file LESSER-GPL;
 * if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package abc.tm.weaving.weaver.tmanalysis;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import soot.Scene;
import soot.SootMethod;
import abc.da.weaving.weaver.depadviceopt.ds.Shadow;
import abc.main.Main;
import abc.tm.weaving.aspectinfo.PerSymbolTMAdviceDecl;
import abc.tm.weaving.aspectinfo.TMGlobalAspectInfo;
import abc.tm.weaving.aspectinfo.TraceMatch;
import abc.tm.weaving.weaver.tmanalysis.Ranking.PotentialFailureGroup;
import abc.tm.weaving.weaver.tmanalysis.util.ShadowsPerTMSplitter;


public class Statistics {
    
	boolean removeIfExists = true;
  
	public void dump(String message, Set<Shadow> inShadows, boolean numberOnly) {
		Set<Shadow> shadows = new HashSet<Shadow>(inShadows);
		
		//remove all non-symbol shadows and all disabled shadows
		for (Iterator<Shadow> iter = shadows.iterator(); iter.hasNext();) {
			Shadow s = iter.next();
			if(!s.isEnabled() || !(s.getAdviceDecl() instanceof PerSymbolTMAdviceDecl)) {
				iter.remove();
			}
		}
		
		//for each tracematch
		Map<TraceMatch, Set<Shadow>> splitSymbolShadows = ShadowsPerTMSplitter.splitSymbolShadows(shadows);		
		for (Map.Entry<TraceMatch, Set<Shadow>> entry : splitSymbolShadows.entrySet()) {
			TraceMatch tm = entry.getKey();
			Set<Shadow> perTMShadows = entry.getValue();

			//compute symbol names
	        Map<SootMethod,String> adviceMethodToSymbol = new HashMap<SootMethod, String>();
	        for (String sym : tm.getSymbols()) {
				SootMethod adviceMethod = tm.getSymbolAdviceMethod(sym);
				adviceMethodToSymbol.put(adviceMethod,sym);
			}
	        Map<Shadow,String> shadowToSymbolName = new HashMap<Shadow, String>();
	        for (Shadow shadow : perTMShadows) {
				SootMethod adviceMethod = shadow.getAdviceDecl().getImpl().getSootMethod();
				String symbol = adviceMethodToSymbol.get(adviceMethod);
				assert symbol!=null;
				shadowToSymbolName.put(shadow, symbol);
			}
			
	        //remove artificial shadows
			for (Iterator<Shadow> iter = perTMShadows.iterator(); iter.hasNext();) {
				Shadow s = iter.next();
				if(shadowToSymbolName.get(s).equals("newDaCapoRun")) {
					iter.remove();
				}
			}
			
			//build qualified tracematch name
			String tmName = tm.getContainerClass().getShortName();
			tmName += "." + tm.getName();

			String bmName = Scene.v().getMainClass().getName();
			String fileName = bmName+"-"+tmName+".pfg";
			if(removeIfExists) {
				new File(fileName).delete();
			}
			
			try {
				PrintWriter out = new PrintWriter(new FileOutputStream(fileName,true));

				Set<Shadow> ppfs = potentialPointOfFailures(perTMShadows);
				List<PotentialFailureGroup> sortedGroups = Ranking.v().rankAndSort(ppfs, tm);
				out.println("Number of potential failure groups "+message+": "+sortedGroups.size());
				if(!numberOnly) {
					for (PotentialFailureGroup potentialFailureGroup : sortedGroups) {
						out.println(potentialFailureGroup);
						out.println();
						out.println();
					}
				}
				out.close();
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			}

		}
		removeIfExists = false;
	}	
	
	private Set<Shadow> potentialPointOfFailures(Set<Shadow> shadows) {
		TMGlobalAspectInfo gai = (TMGlobalAspectInfo) Main.v().getAbcExtension().getGlobalAspectInfo();

		Set<Shadow> PPFs = new HashSet<Shadow>();
		
    	Map<SootMethod,String> adviceMethodToSymbol = new HashMap<SootMethod, String>();
    	Map<SootMethod,TraceMatch> adviceMethodToTraceMatch = new HashMap<SootMethod, TraceMatch>();
    	for(TraceMatch tm : gai.getTraceMatches()) {	        
	        for (String sym : tm.getSymbols()) {
				SootMethod adviceMethod = tm.getSymbolAdviceMethod(sym);
				adviceMethodToSymbol.put(adviceMethod,sym);
				adviceMethodToTraceMatch.put(adviceMethod,tm);
			}
    	}
        Set<Shadow> allShadows = new HashSet<Shadow>(shadows);
        for (Shadow s : allShadows) {
			if(s.isEnabled() && (s.getAdviceDecl() instanceof PerSymbolTMAdviceDecl)) {
				SootMethod adviceMethod = s.getAdviceDecl().getImpl().getSootMethod();
				String symbol = adviceMethodToSymbol.get(adviceMethod);
				assert symbol!=null;
				if(adviceMethodToTraceMatch.get(adviceMethod).getFinalSymbols().
						contains(adviceMethodToSymbol.get(adviceMethod))) {
					PPFs.add(s);
				}
			}
		}
                
        return PPFs;
	}

	//singleton pattern
	
	protected static Statistics instance;

    private Statistics() { }

    public static Statistics v() {
		if(instance==null) {
			instance = new Statistics();
		}
		return instance;		
	}
	
	/**
	 * Frees the singleton object. 
	 */
	public static void reset() {
		instance = null;
	}


}
